@@ -21,8 +21,10 @@
 package org.apache.bookkeeper.bookie.storage.ldb;
 
 import static com.google.common.base.Preconditions.checkArgument;
+import static com.google.common.base.Preconditions.checkState;
 
 import com.google.common.annotations.VisibleForTesting;
+import com.google.common.collect.ImmutableMap;
 import com.google.common.collect.Lists;
 import com.google.protobuf.ByteString;
 
@@ -35,7 +37,9 @@ import io.netty.util.concurrent.DefaultThreadFactory;
 import java.io.File;
 import java.io.IOException;
 import java.util.Collections;
+import java.util.EnumSet;
 import java.util.List;
+import java.util.Map;
 import java.util.PrimitiveIterator.OfLong;
 import java.util.concurrent.CopyOnWriteArrayList;
 import java.util.concurrent.ExecutorService;
@@ -70,8 +74,10 @@ import org.apache.bookkeeper.common.util.Watcher;
 import org.apache.bookkeeper.conf.ServerConfiguration;
 import org.apache.bookkeeper.meta.LedgerManager;
 import org.apache.bookkeeper.proto.BookieProtocol;
+import org.apache.bookkeeper.stats.Counter;
 import org.apache.bookkeeper.stats.OpStatsLogger;
 import org.apache.bookkeeper.stats.StatsLogger;
+import org.apache.bookkeeper.stats.ThreadRegistry;
 import org.apache.bookkeeper.util.MathUtils;
 import org.apache.bookkeeper.util.collections.ConcurrentLongHashMap;
 import org.apache.commons.lang.mutable.MutableLong;
@@ -110,7 +116,9 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
     protected final AtomicBoolean hasFlushBeenTriggered = new AtomicBoolean(false);
     private final AtomicBoolean isFlushOngoing = new AtomicBoolean(false);
 
-    private final ExecutorService executor = Executors.newSingleThreadExecutor(new DefaultThreadFactory("db-storage"));
+    private static String dbStoragerExecutorName = "db-storage";
+    private final ExecutorService executor = Executors.newSingleThreadExecutor(
+            new DefaultThreadFactory(dbStoragerExecutorName));
 
     // Executor used to for db index cleanup
     private final ScheduledExecutorService cleanupExecutor = Executors
@@ -119,7 +127,7 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
     private final CopyOnWriteArrayList<LedgerDeletionListener> ledgerDeletionListeners = Lists
             .newCopyOnWriteArrayList();
 
-    private final CheckpointSource checkpointSource;
+    private CheckpointSource checkpointSource = CheckpointSource.DEFAULT;
     private Checkpoint lastCheckpoint = Checkpoint.MIN;
 
     private final long writeCacheMaxSize;
@@ -130,36 +138,31 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
 
     private final DbLedgerStorageStats dbLedgerStorageStats;
 
-    static final String READ_AHEAD_CACHE_BATCH_SIZE = "dbStorage_readAheadCacheBatchSize";
-    private static final int DEFAULT_READ_AHEAD_CACHE_BATCH_SIZE = 100;
-
     private static final long DEFAULT_MAX_THROTTLE_TIME_MILLIS = TimeUnit.SECONDS.toMillis(10);
 
     private final long maxReadAheadBytesSize;
 
-    private final boolean singleLedgerDirs;
+    private final Counter flushExecutorTime;
 
     public SingleDirectoryDbLedgerStorage(ServerConfiguration conf, LedgerManager ledgerManager,
-            LedgerDirsManager ledgerDirsManager, LedgerDirsManager indexDirsManager, StateManager stateManager,
-            CheckpointSource checkpointSource, Checkpointer checkpointer, StatsLogger statsLogger,
-            ByteBufAllocator allocator, ScheduledExecutorService gcExecutor, long writeCacheSize, long readCacheSize)
-            throws IOException {
-
+            LedgerDirsManager ledgerDirsManager, LedgerDirsManager indexDirsManager, StatsLogger statsLogger,
+            ByteBufAllocator allocator, ScheduledExecutorService gcExecutor, long writeCacheSize, long readCacheSize,
+            int readAheadCacheBatchSize) throws IOException {
         checkArgument(ledgerDirsManager.getAllLedgerDirs().size() == 1,
                 "Db implementation only allows for one storage dir");
 
         String baseDir = ledgerDirsManager.getAllLedgerDirs().get(0).toString();
         log.info("Creating single directory db ledger storage on {}", baseDir);
 
+        StatsLogger ledgerDirStatsLogger = statsLogger.scopeLabel("ledgerDir",
+                ledgerDirsManager.getAllLedgerDirs().get(0).getPath());
+
         this.writeCacheMaxSize = writeCacheSize;
         this.writeCache = new WriteCache(allocator, writeCacheMaxSize / 2);
         this.writeCacheBeingFlushed = new WriteCache(allocator, writeCacheMaxSize / 2);
-        this.singleLedgerDirs = conf.getLedgerDirs().length == 1;
-
-        this.checkpointSource = checkpointSource;
 
         readCacheMaxSize = readCacheSize;
-        readAheadCacheBatchSize = conf.getInt(READ_AHEAD_CACHE_BATCH_SIZE, DEFAULT_READ_AHEAD_CACHE_BATCH_SIZE);
+        this.readAheadCacheBatchSize = readAheadCacheBatchSize;
 
         // Do not attempt to perform read-ahead more than half the total size of the cache
         maxReadAheadBytesSize = readCacheMaxSize / 2;
@@ -170,36 +173,58 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
 
         readCache = new ReadCache(allocator, readCacheMaxSize);
 
-        ledgerIndex = new LedgerMetadataIndex(conf, KeyValueStorageRocksDB.factory, baseDir, statsLogger);
-        entryLocationIndex = new EntryLocationIndex(conf, KeyValueStorageRocksDB.factory, baseDir, statsLogger);
+        ledgerIndex = new LedgerMetadataIndex(conf, KeyValueStorageRocksDB.factory, baseDir, ledgerDirStatsLogger);
+        entryLocationIndex = new EntryLocationIndex(conf,
+                KeyValueStorageRocksDB.factory, baseDir, ledgerDirStatsLogger);
 
-        transientLedgerInfoCache = new ConcurrentLongHashMap<>(16 * 1024,
-                Runtime.getRuntime().availableProcessors() * 2);
+        transientLedgerInfoCache = ConcurrentLongHashMap.<TransientLedgerInfo>newBuilder()
+                .expectedItems(16 * 1024)
+                .concurrencyLevel(Runtime.getRuntime().availableProcessors() * 2)
+                .build();
         cleanupExecutor.scheduleAtFixedRate(this::cleanupStaleTransientLedgerInfo,
                 TransientLedgerInfo.LEDGER_INFO_CACHING_TIME_MINUTES,
                 TransientLedgerInfo.LEDGER_INFO_CACHING_TIME_MINUTES, TimeUnit.MINUTES);
 
         entryLogger = new EntryLogger(conf, ledgerDirsManager, null, statsLogger, allocator);
-        gcThread = new GarbageCollectorThread(conf, ledgerManager, this, statsLogger);
+        gcThread = new GarbageCollectorThread(conf, ledgerManager, ledgerDirsManager, this, statsLogger);
 
         dbLedgerStorageStats = new DbLedgerStorageStats(
-            statsLogger,
+                ledgerDirStatsLogger,
             () -> writeCache.size() + writeCacheBeingFlushed.size(),
             () -> writeCache.count() + writeCacheBeingFlushed.count(),
             () -> readCache.size(),
             () -> readCache.count()
         );
+
+        flushExecutorTime = ledgerDirStatsLogger.getThreadScopedCounter("db-storage-thread-time");
+
+        executor.submit(() -> {
+            ThreadRegistry.register(dbStoragerExecutorName, 0);
+            // ensure the metric gets registered on start-up as this thread only executes
+            // when the write cache is full which may not happen or not for a long time
+            flushExecutorTime.add(0);
+        });
+
         ledgerDirsManager.addLedgerDirsListener(getLedgerDirsListener());
     }
 
     @Override
     public void initialize(ServerConfiguration conf, LedgerManager ledgerManager, LedgerDirsManager ledgerDirsManager,
-            LedgerDirsManager indexDirsManager, StateManager stateManager, CheckpointSource checkpointSource,
-            Checkpointer checkpointer, StatsLogger statsLogger,
+            LedgerDirsManager indexDirsManager, StatsLogger statsLogger,
             ByteBufAllocator allocator) throws IOException {
         /// Initialized in constructor
     }
 
+    @Override
+    public void setStateManager(StateManager stateManager) { }
+
+    @Override
+    public void setCheckpointSource(CheckpointSource checkpointSource) {
+        this.checkpointSource = checkpointSource;
+    }
+    @Override
+    public void setCheckpointer(Checkpointer checkpointer) { }
+
     /**
      * Evict all the ledger info object that were not used recently.
      */
@@ -224,6 +249,11 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
         gcThread.enableForceGC();
     }
 
+    @Override
+    public void forceGC(Boolean forceMajor, Boolean forceMinor) {
+        gcThread.enableForceGC(forceMajor, forceMinor);
+    }
+
     @Override
     public boolean isInForceGC() {
         return gcThread.isInForceGC();
@@ -268,11 +298,59 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
     }
 
     @Override
-    public boolean isFenced(long ledgerId) throws IOException {
+    public boolean entryExists(long ledgerId, long entryId) throws IOException, BookieException {
+        if (entryId == BookieProtocol.LAST_ADD_CONFIRMED) {
+            return false;
+        }
+
+        // We need to try to read from both write caches, since recent entries could be found in either of the two. The
+        // write caches are already thread safe on their own, here we just need to make sure we get references to both
+        // of them. Using an optimistic lock since the read lock is always free, unless we're swapping the caches.
+        long stamp = writeCacheRotationLock.tryOptimisticRead();
+        WriteCache localWriteCache = writeCache;
+        WriteCache localWriteCacheBeingFlushed = writeCacheBeingFlushed;
+        if (!writeCacheRotationLock.validate(stamp)) {
+            // Fallback to regular read lock approach
+            stamp = writeCacheRotationLock.readLock();
+            try {
+                localWriteCache = writeCache;
+                localWriteCacheBeingFlushed = writeCacheBeingFlushed;
+            } finally {
+                writeCacheRotationLock.unlockRead(stamp);
+            }
+        }
+
+        boolean inCache = localWriteCache.hasEntry(ledgerId, entryId)
+             || localWriteCacheBeingFlushed.hasEntry(ledgerId, entryId)
+             || readCache.hasEntry(ledgerId, entryId);
+
+        if (inCache) {
+            return true;
+        }
+
+        // Read from main storage
+        long entryLocation = entryLocationIndex.getLocation(ledgerId, entryId);
+        if (entryLocation != 0) {
+            return true;
+        }
+
+        // Only a negative result while in limbo equates to unknown
+        throwIfLimbo(ledgerId);
+
+        return false;
+    }
+
+    @Override
+    public boolean isFenced(long ledgerId) throws IOException, BookieException {
+        if (log.isDebugEnabled()) {
+            log.debug("isFenced. ledger: {}", ledgerId);
+        }
+
         boolean isFenced = ledgerIndex.get(ledgerId).getFenced();
 
-        if (log.isDebugEnabled()) {
-            log.debug("ledger: {}, isFenced: {}.", ledgerId, isFenced);
+        // Only a negative result while in limbo equates to unknown
+        if (!isFenced) {
+            throwIfLimbo(ledgerId);
         }
 
         return isFenced;
@@ -366,10 +444,13 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
                 // Trigger an early flush in background
                 log.info("Write cache is full, triggering flush");
                 executor.execute(() -> {
+                        long startTime = System.nanoTime();
                         try {
                             flush();
                         } catch (IOException e) {
                             log.error("Error during flush", e);
+                        } finally {
+                            flushExecutorTime.add(MathUtils.elapsedNanos(startTime));
                         }
                     });
             }
@@ -401,8 +482,19 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
     }
 
     @Override
-    public ByteBuf getEntry(long ledgerId, long entryId) throws IOException {
+    public ByteBuf getEntry(long ledgerId, long entryId) throws IOException, BookieException {
         long startTime = MathUtils.nowInNano();
+        try {
+            ByteBuf entry = doGetEntry(ledgerId, entryId);
+            recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+            return entry;
+        } catch (IOException e) {
+            recordFailedEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+            throw e;
+        }
+    }
+
+    private ByteBuf doGetEntry(long ledgerId, long entryId) throws IOException, BookieException {
         if (log.isDebugEnabled()) {
             log.debug("Get Entry: {}@{}", ledgerId, entryId);
         }
@@ -431,50 +523,48 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
         // First try to read from the write cache of recent entries
         ByteBuf entry = localWriteCache.get(ledgerId, entryId);
         if (entry != null) {
-            recordSuccessfulEvent(dbLedgerStorageStats.getReadCacheHitStats(), startTime);
-            recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+            dbLedgerStorageStats.getWriteCacheHitCounter().inc();
             return entry;
         }
 
         // If there's a flush going on, the entry might be in the flush buffer
         entry = localWriteCacheBeingFlushed.get(ledgerId, entryId);
         if (entry != null) {
-            recordSuccessfulEvent(dbLedgerStorageStats.getReadCacheHitStats(), startTime);
-            recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+            dbLedgerStorageStats.getWriteCacheHitCounter().inc();
             return entry;
         }
 
+        dbLedgerStorageStats.getWriteCacheMissCounter().inc();
+
         // Try reading from read-ahead cache
         entry = readCache.get(ledgerId, entryId);
         if (entry != null) {
-            recordSuccessfulEvent(dbLedgerStorageStats.getReadCacheHitStats(), startTime);
-            recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+            dbLedgerStorageStats.getReadCacheHitCounter().inc();
             return entry;
         }
 
+        dbLedgerStorageStats.getReadCacheMissCounter().inc();
+
         // Read from main storage
         long entryLocation;
         long locationIndexStartNano = MathUtils.nowInNano();
         try {
             entryLocation = entryLocationIndex.getLocation(ledgerId, entryId);
             if (entryLocation == 0) {
-                recordFailedEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+                // Only a negative result while in limbo equates to unknown
+                throwIfLimbo(ledgerId);
+
                 throw new NoEntryException(ledgerId, entryId);
             }
         } finally {
-            dbLedgerStorageStats.getReadFromLocationIndexTime().registerSuccessfulEvent(
-                MathUtils.elapsedNanos(locationIndexStartNano), TimeUnit.NANOSECONDS);
+            dbLedgerStorageStats.getReadFromLocationIndexTime().add(MathUtils.elapsedNanos(locationIndexStartNano));
         }
 
         long readEntryStartNano = MathUtils.nowInNano();
         try {
             entry = entryLogger.readEntry(ledgerId, entryId, entryLocation);
-        } catch (NoEntryException e) {
-            recordFailedEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
-            throw e;
         } finally {
-            dbLedgerStorageStats.getReadFromEntryLogTime().registerSuccessfulEvent(
-                MathUtils.elapsedNanos(readEntryStartNano), TimeUnit.NANOSECONDS);
+            dbLedgerStorageStats.getReadFromEntryLogTime().add(MathUtils.elapsedNanos(readEntryStartNano));
         }
 
         readCache.put(ledgerId, entryId, entry);
@@ -483,18 +573,18 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
         long nextEntryLocation = entryLocation + 4 /* size header */ + entry.readableBytes();
         fillReadAheadCache(ledgerId, entryId + 1, nextEntryLocation);
 
-        recordSuccessfulEvent(dbLedgerStorageStats.getReadCacheMissStats(), startTime);
-        recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
         return entry;
     }
 
     private void fillReadAheadCache(long orginalLedgerId, long firstEntryId, long firstEntryLocation) {
+        long readAheadStartNano = MathUtils.nowInNano();
+        int count = 0;
+        long size = 0;
+
         try {
             long firstEntryLogId = (firstEntryLocation >> 32);
             long currentEntryLogId = firstEntryLogId;
             long currentEntryLocation = firstEntryLocation;
-            int count = 0;
-            long size = 0;
 
             while (count < readAheadCacheBatchSize
                     && size < maxReadAheadBytesSize
@@ -524,18 +614,19 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
                     entry.release();
                 }
             }
-
-            dbLedgerStorageStats.getReadAheadBatchCountStats().registerSuccessfulValue(count);
-            dbLedgerStorageStats.getReadAheadBatchSizeStats().registerSuccessfulValue(size);
         } catch (Exception e) {
             if (log.isDebugEnabled()) {
                 log.debug("Exception during read ahead for ledger: {}: e", orginalLedgerId, e);
             }
+        } finally {
+            dbLedgerStorageStats.getReadAheadBatchCountStats().registerSuccessfulValue(count);
+            dbLedgerStorageStats.getReadAheadBatchSizeStats().registerSuccessfulValue(size);
+            dbLedgerStorageStats.getReadAheadTime().add(MathUtils.elapsedNanos(readAheadStartNano));
         }
     }
 
-    public ByteBuf getLastEntry(long ledgerId) throws IOException {
-        long startTime = MathUtils.nowInNano();
+    public ByteBuf getLastEntry(long ledgerId) throws IOException, BookieException {
+        throwIfLimbo(ledgerId);
 
         long stamp = writeCacheRotationLock.readLock();
         try {
@@ -552,8 +643,7 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
                     }
                 }
 
-                recordSuccessfulEvent(dbLedgerStorageStats.getReadCacheHitStats(), startTime);
-                recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+                dbLedgerStorageStats.getWriteCacheHitCounter().inc();
                 return entry;
             }
 
@@ -569,25 +659,28 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
                     }
                 }
 
-                recordSuccessfulEvent(dbLedgerStorageStats.getReadCacheHitStats(), startTime);
-                recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+                dbLedgerStorageStats.getWriteCacheHitCounter().inc();
                 return entry;
             }
         } finally {
             writeCacheRotationLock.unlockRead(stamp);
         }
 
+        dbLedgerStorageStats.getWriteCacheMissCounter().inc();
+
         // Search the last entry in storage
+        long locationIndexStartNano = MathUtils.nowInNano();
         long lastEntryId = entryLocationIndex.getLastEntryInLedger(ledgerId);
         if (log.isDebugEnabled()) {
             log.debug("Found last entry for ledger {} in db: {}", ledgerId, lastEntryId);
         }
 
         long entryLocation = entryLocationIndex.getLocation(ledgerId, lastEntryId);
-        ByteBuf content = entryLogger.readEntry(ledgerId, lastEntryId, entryLocation);
+        dbLedgerStorageStats.getReadFromLocationIndexTime().add(MathUtils.elapsedNanos(locationIndexStartNano));
 
-        recordSuccessfulEvent(dbLedgerStorageStats.getReadCacheMissStats(), startTime);
-        recordSuccessfulEvent(dbLedgerStorageStats.getReadEntryStats(), startTime);
+        long readEntryStartNano = MathUtils.nowInNano();
+        ByteBuf content = entryLogger.readEntry(ledgerId, lastEntryId, entryLocation);
+        dbLedgerStorageStats.getReadFromEntryLogTime().add(MathUtils.elapsedNanos(readEntryStartNano));
         return content;
     }
 
@@ -612,8 +705,10 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
 
         // Only a single flush operation can happen at a time
         flushMutex.lock();
-
         try {
+            if (writeCache.isEmpty()) {
+                return;
+            }
             // Swap the write cache so that writes can continue to happen while the flush is
             // ongoing
             swapWriteCache();
@@ -637,17 +732,37 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
                 }
             });
 
+            long entryLoggerStart = MathUtils.nowInNano();
             entryLogger.flush();
+            recordSuccessfulEvent(dbLedgerStorageStats.getFlushEntryLogStats(), entryLoggerStart);
 
-            long batchFlushStarTime = System.nanoTime();
+            long batchFlushStartTime = MathUtils.nowInNano();
             batch.flush();
             batch.close();
+            recordSuccessfulEvent(dbLedgerStorageStats.getFlushLocationIndexStats(), batchFlushStartTime);
             if (log.isDebugEnabled()) {
                 log.debug("DB batch flushed time : {} s",
-                        MathUtils.elapsedNanos(batchFlushStarTime) / (double) TimeUnit.SECONDS.toNanos(1));
+                        MathUtils.elapsedNanos(batchFlushStartTime) / (double) TimeUnit.SECONDS.toNanos(1));
             }
 
+            long ledgerIndexStartTime = MathUtils.nowInNano();
             ledgerIndex.flush();
+            recordSuccessfulEvent(dbLedgerStorageStats.getFlushLedgerIndexStats(), ledgerIndexStartTime);
+
+            cleanupExecutor.execute(() -> {
+                // There can only be one single cleanup task running because the cleanupExecutor
+                // is single-threaded
+                try {
+                    if (log.isDebugEnabled()) {
+                        log.debug("Removing deleted ledgers from db indexes");
+                    }
+
+                    entryLocationIndex.removeOffsetFromDeletedLedgers();
+                    ledgerIndex.removeDeletedLedgers();
+                } catch (Throwable t) {
+                    log.warn("Failed to cleanup db indexes", t);
+                }
+            });
 
             lastCheckpoint = thisCheckpoint;
 
@@ -664,28 +779,15 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
             recordSuccessfulEvent(dbLedgerStorageStats.getFlushStats(), startTime);
             dbLedgerStorageStats.getFlushSizeStats().registerSuccessfulValue(sizeToFlush);
         } catch (IOException e) {
+            recordFailedEvent(dbLedgerStorageStats.getFlushStats(), startTime);
             // Leave IOExecption as it is
             throw e;
         } catch (RuntimeException e) {
+            recordFailedEvent(dbLedgerStorageStats.getFlushStats(), startTime);
             // Wrap unchecked exceptions
             throw new IOException(e);
         } finally {
             try {
-                cleanupExecutor.execute(() -> {
-                    // There can only be one single cleanup task running because the cleanupExecutor
-                    // is single-threaded
-                    try {
-                        if (log.isDebugEnabled()) {
-                            log.debug("Removing deleted ledgers from db indexes");
-                        }
-
-                        entryLocationIndex.removeOffsetFromDeletedLedgers();
-                        ledgerIndex.removeDeletedLedgers();
-                    } catch (Throwable t) {
-                        log.warn("Failed to cleanup db indexes", t);
-                    }
-                });
-
                 isFlushOngoing.set(false);
             } finally {
                 flushMutex.unlock();
@@ -720,9 +822,7 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
     public void flush() throws IOException {
         Checkpoint cp = checkpointSource.newCheckpoint();
         checkpoint(cp);
-        if (singleLedgerDirs) {
-            checkpointSource.checkpointComplete(cp, true);
-        }
+        checkpointSource.checkpointComplete(cp, true);
     }
 
     @Override
@@ -760,30 +860,9 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
 
     @Override
     public void updateEntriesLocations(Iterable<EntryLocation> locations) throws IOException {
-        // Before updating the DB with the new location for the compacted entries, we need to
-        // make sure that there is no ongoing flush() operation.
-        // If there were a flush, we could have the following situation, which is highly
-        // unlikely though possible:
-        // 1. Flush operation has written the write-cache content into entry-log files
-        // 2. The DB location index is not yet updated
-        // 3. Compaction is triggered and starts compacting some of the recent files
-        // 4. Compaction will write the "new location" into the DB
-        // 5. The pending flush() will overwrite the DB with the "old location", pointing
-        //    to a file that no longer exists
-        //
-        // To avoid this race condition, we need that all the entries that are potentially
-        // included in the compaction round to have all the indexes already flushed into
-        // the DB.
-        // The easiest lightweight way to achieve this is to wait for any pending
-        // flush operation to be completed before updating the index with the compacted
-        // entries, by blocking on the flushMutex.
-        flushMutex.lock();
-        flushMutex.unlock();
+        // Trigger a flush to have all the entries being compacted in the db storage
+        flush();
 
-        // We don't need to keep the flush mutex locked here while updating the DB.
-        // It's fine to have a concurrent flush operation at this point, because we
-        // know that none of the entries being flushed was included in the compaction
-        // round that we are dealing with.
         entryLocationIndex.updateLocations(locations);
     }
 
@@ -793,7 +872,9 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
     }
 
     @Override
-    public long getLastAddConfirmed(long ledgerId) throws IOException {
+    public long getLastAddConfirmed(long ledgerId) throws IOException, BookieException {
+        throwIfLimbo(ledgerId);
+
         TransientLedgerInfo ledgerInfo = transientLedgerInfoCache.get(ledgerId);
         long lac = null != ledgerInfo ? ledgerInfo.getLastAddConfirmed() : TransientLedgerInfo.NOT_ASSIGNED_LAC;
         if (lac == TransientLedgerInfo.NOT_ASSIGNED_LAC) {
@@ -831,7 +912,8 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
     }
 
     @Override
-    public ByteBuf getExplicitLac(long ledgerId) throws IOException {
+    public ByteBuf getExplicitLac(long ledgerId) throws IOException, BookieException {
+        throwIfLimbo(ledgerId);
         if (log.isDebugEnabled()) {
             log.debug("getExplicitLac ledger {}", ledgerId);
         }
@@ -932,22 +1014,6 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
         logger.registerFailedEvent(MathUtils.elapsedNanos(startTimeNanos), TimeUnit.NANOSECONDS);
     }
 
-    long getWriteCacheSize() {
-        return writeCache.size() + writeCacheBeingFlushed.size();
-    }
-
-    long getWriteCacheCount() {
-        return writeCache.count() + writeCacheBeingFlushed.count();
-    }
-
-    long getReadCacheSize() {
-        return readCache.size();
-    }
-
-    long getReadCacheCount() {
-        return readCache.count();
-    }
-
     @Override
     public List<GarbageCollectionStatus> getGarbageCollectionStatus() {
         return Collections.singletonList(gcThread.getGarbageCollectionStatus());
@@ -1025,4 +1091,91 @@ public class SingleDirectoryDbLedgerStorage implements CompactableLedgerStorage
             }
         };
     }
+
+    @Override
+    public void setLimboState(long ledgerId) throws IOException {
+        if (log.isDebugEnabled()) {
+            log.debug("setLimboState. ledger: {}", ledgerId);
+        }
+        ledgerIndex.setLimbo(ledgerId);
+    }
+
+    @Override
+    public boolean hasLimboState(long ledgerId) throws IOException {
+        if (log.isDebugEnabled()) {
+            log.debug("hasLimboState. ledger: {}", ledgerId);
+        }
+        return ledgerIndex.get(ledgerId).getLimbo();
+    }
+
+    @Override
+    public void clearLimboState(long ledgerId) throws IOException {
+        if (log.isDebugEnabled()) {
+            log.debug("clearLimboState. ledger: {}", ledgerId);
+        }
+        ledgerIndex.clearLimbo(ledgerId);
+    }
+
+    private void throwIfLimbo(long ledgerId) throws IOException, BookieException {
+        if (hasLimboState(ledgerId)) {
+            if (log.isDebugEnabled()) {
+                log.debug("Accessing ledger({}) in limbo state, throwing exception", ledgerId);
+            }
+            throw BookieException.create(BookieException.Code.DataUnknownException);
+        }
+    }
+
+    /**
+     * Mapping of enums to bitmaps. The bitmaps must not overlap so that we can
+     * do bitwise operations on them.
+     */
+    private static final Map<StorageState, Integer> stateBitmaps = ImmutableMap.of(
+            StorageState.NEEDS_INTEGRITY_CHECK, 0x00000001);
+
+    @Override
+    public EnumSet<StorageState> getStorageStateFlags() throws IOException {
+        int flags = ledgerIndex.getStorageStateFlags();
+        EnumSet<StorageState> flagsEnum = EnumSet.noneOf(StorageState.class);
+        for (Map.Entry<StorageState, Integer> e : stateBitmaps.entrySet()) {
+            int value = e.getValue();
+            if ((flags & value) == value) {
+                flagsEnum.add(e.getKey());
+            }
+            flags = flags & ~value;
+        }
+        checkState(flags == 0, "Unknown storage state flag found " + flags);
+        return flagsEnum;
+    }
+
+    @Override
+    public void setStorageStateFlag(StorageState flag) throws IOException {
+        checkArgument(stateBitmaps.containsKey(flag), "Unsupported flag " + flag);
+        int flagInt = stateBitmaps.get(flag);
+        while (true) {
+            int curFlags = ledgerIndex.getStorageStateFlags();
+            int newFlags = curFlags | flagInt;
+            if (ledgerIndex.setStorageStateFlags(curFlags, newFlags)) {
+                return;
+            } else {
+                log.info("Conflict updating storage state flags {} -> {}, retrying",
+                        curFlags, newFlags);
+            }
+        }
+    }
+
+    @Override
+    public void clearStorageStateFlag(StorageState flag) throws IOException {
+        checkArgument(stateBitmaps.containsKey(flag), "Unsupported flag " + flag);
+        int flagInt = stateBitmaps.get(flag);
+        while (true) {
+            int curFlags = ledgerIndex.getStorageStateFlags();
+            int newFlags = curFlags & ~flagInt;
+            if (ledgerIndex.setStorageStateFlags(curFlags, newFlags)) {
+                return;
+            } else {
+                log.info("Conflict updating storage state flags {} -> {}, retrying",
+                        curFlags, newFlags);
+            }
+        }
+    }
 }
