@@ -157,12 +157,12 @@ public class BookieRequestProcessor implements RequestProcessor {
             }
             this.longPollThreadPool = createExecutor(
                 numThreads,
-                "BookieLongPollThread-" + serverCfg.getBookiePort(),
+                "BookieLongPollThread",
                 OrderedExecutor.NO_TASK_LIMIT, statsLogger);
         }
         this.highPriorityThreadPool = createExecutor(
                 this.serverCfg.getNumHighPriorityWorkerThreads(),
-                "BookieHighPriorityThread-" + serverCfg.getBookiePort(),
+                "BookieHighPriorityThread",
                 OrderedExecutor.NO_TASK_LIMIT, statsLogger);
         this.shFactory = shFactory;
         if (shFactory != null) {
@@ -290,6 +290,7 @@ public class BookieRequestProcessor implements RequestProcessor {
                     .preserveMdcForTaskExecution(serverCfg.getPreserveMdcForTaskExecution())
                     .statsLogger(statsLogger)
                     .maxTasksInQueue(maxTasksInQueue)
+                    .enableThreadScopedMetrics(true)
                     .build();
         }
     }
@@ -443,6 +444,7 @@ public class BookieRequestProcessor implements RequestProcessor {
                     LOG.debug("Failed to process request to add entry at {}:{}. Too many pending requests",
                               r.getAddRequest().getLedgerId(), r.getAddRequest().getEntryId());
                 }
+                getRequestStats().getAddEntryRejectedCounter().inc();
                 BookkeeperProtocol.AddResponse.Builder addResponse = BookkeeperProtocol.AddResponse.newBuilder()
                         .setLedgerId(r.getAddRequest().getLedgerId())
                         .setEntryId(r.getAddRequest().getEntryId())
@@ -530,6 +532,7 @@ public class BookieRequestProcessor implements RequestProcessor {
                     LOG.debug("Failed to process request to read entry at {}:{}. Too many pending requests",
                               r.getReadRequest().getLedgerId(), r.getReadRequest().getEntryId());
                 }
+                getRequestStats().getReadEntryRejectedCounter().inc();
                 BookkeeperProtocol.ReadResponse.Builder readResponse = BookkeeperProtocol.ReadResponse.newBuilder()
                     .setLedgerId(r.getReadRequest().getLedgerId())
                     .setEntryId(r.getReadRequest().getEntryId())
@@ -540,7 +543,6 @@ public class BookieRequestProcessor implements RequestProcessor {
                     .setReadResponse(readResponse);
                 BookkeeperProtocol.Response resp = response.build();
                 read.sendResponse(readResponse.getStatus(), resp, requestStats.getReadRequestStats());
-                onReadRequestFinish();
             }
         }
     }
@@ -638,6 +640,7 @@ public class BookieRequestProcessor implements RequestProcessor {
                     LOG.debug("Failed to process request to add entry at {}:{}. Too many pending requests", r.ledgerId,
                             r.entryId);
                 }
+                getRequestStats().getAddEntryRejectedCounter().inc();
 
                 write.sendWriteReqResponse(
                     BookieProtocol.ETOOMANYREQUESTS,
@@ -672,12 +675,11 @@ public class BookieRequestProcessor implements RequestProcessor {
                     LOG.debug("Failed to process request to read entry at {}:{}. Too many pending requests", r.ledgerId,
                             r.entryId);
                 }
-
+                getRequestStats().getReadEntryRejectedCounter().inc();
                 read.sendResponse(
                     BookieProtocol.ETOOMANYREQUESTS,
                     ResponseBuilder.buildErrorResponse(BookieProtocol.ETOOMANYREQUESTS, r),
                     requestStats.getReadRequestStats());
-                onReadRequestFinish();
             }
         }
     }
