@@ -108,7 +108,7 @@ public abstract class BookKeeperClusterTestCase {
 
     // BookKeeper related variables
     protected final TmpDirs tmpDirs = new TmpDirs();
-    private final List<ServerTester> servers = new LinkedList<>();
+    protected final List<ServerTester> servers = new LinkedList<>();
 
     protected int numBookies;
     protected BookKeeperTestClient bkc;
@@ -835,7 +835,7 @@ public abstract class BookKeeperClusterTestCase {
 
         private AutoRecoveryMain autoRecovery;
 
-        ServerTester(ServerConfiguration conf) throws Exception {
+        public ServerTester(ServerConfiguration conf) throws Exception {
             this.conf = conf;
             provider = new TestStatsProvider();
 
@@ -879,7 +879,7 @@ public abstract class BookKeeperClusterTestCase {
             autoRecovery = null;
         }
 
-        ServerTester(ServerConfiguration conf, Bookie b) throws Exception {
+        public ServerTester(ServerConfiguration conf, Bookie b) throws Exception {
             this.conf = conf;
             provider = new TestStatsProvider();
 
@@ -897,20 +897,20 @@ public abstract class BookKeeperClusterTestCase {
             autoRecovery = null;
         }
 
-        void startAutoRecovery() throws Exception {
+        public void startAutoRecovery() throws Exception {
             LOG.debug("Starting Auditor Recovery for the bookie: {}", address);
             autoRecovery = new AutoRecoveryMain(conf);
             autoRecovery.start();
         }
 
-        void stopAutoRecovery() {
+        public void stopAutoRecovery() {
             if (autoRecovery != null) {
                 LOG.debug("Shutdown Auditor Recovery for the bookie: {}", address);
                 autoRecovery.shutdown();
             }
         }
 
-        Auditor getAuditor() {
+        public Auditor getAuditor() {
             if (autoRecovery != null) {
                 return autoRecovery.getAuditor();
             } else {
@@ -918,7 +918,7 @@ public abstract class BookKeeperClusterTestCase {
             }
         }
 
-        ReplicationWorker getReplicationWorker() {
+        public ReplicationWorker getReplicationWorker() {
             if (autoRecovery != null) {
                 return autoRecovery.getReplicationWorker();
             } else {
@@ -926,7 +926,7 @@ public abstract class BookKeeperClusterTestCase {
             }
         }
 
-        ServerConfiguration getConfiguration() {
+        public ServerConfiguration getConfiguration() {
             return conf;
         }
 
@@ -934,15 +934,15 @@ public abstract class BookKeeperClusterTestCase {
             return server;
         }
 
-        TestStatsProvider getStatsProvider() {
+        public TestStatsProvider getStatsProvider() {
             return provider;
         }
 
-        BookieSocketAddress getAddress() {
+        public BookieSocketAddress getAddress() {
             return address;
         }
 
-        void shutdown() throws Exception {
+        public void shutdown() throws Exception {
             server.shutdown();
 
             if (ledgerManager != null) {
