@@ -21,51 +21,35 @@ package org.apache.bookkeeper.tools.cli.commands.client;
 import static org.mockito.ArgumentMatchers.any;
 import static org.mockito.ArgumentMatchers.anyLong;
 import static org.mockito.ArgumentMatchers.eq;
+import static org.mockito.Mockito.mock;
 import static org.mockito.Mockito.times;
 import static org.mockito.Mockito.verify;
-import static org.powermock.api.mockito.PowerMockito.mock;
-import static org.powermock.api.mockito.PowerMockito.when;
-import static org.powermock.api.mockito.PowerMockito.whenNew;
+import static org.mockito.Mockito.when;
 
 import java.io.File;
 import java.io.IOException;
 import java.util.Optional;
 import java.util.concurrent.CompletableFuture;
-import java.util.function.Function;
 import org.apache.bookkeeper.client.api.LedgerMetadata;
-import org.apache.bookkeeper.conf.ServerConfiguration;
 import org.apache.bookkeeper.meta.LedgerManager;
 import org.apache.bookkeeper.meta.LedgerManagerFactory;
 import org.apache.bookkeeper.meta.LedgerMetadataSerDe;
-import org.apache.bookkeeper.meta.MetadataDrivers;
 import org.apache.bookkeeper.tools.cli.helpers.BookieCommandTestBase;
 import org.apache.bookkeeper.versioning.Version;
 import org.apache.bookkeeper.versioning.Versioned;
 import org.junit.Assert;
 import org.junit.Test;
-import org.junit.rules.TemporaryFolder;
-import org.junit.runner.RunWith;
-import org.mockito.Mock;
-import org.powermock.api.mockito.PowerMockito;
-import org.powermock.core.classloader.annotations.PrepareForTest;
-import org.powermock.modules.junit4.PowerMockRunner;
 
 /**
  * Unit test for {@link LedgerMetaDataCommand}.
  */
-@RunWith(PowerMockRunner.class)
-@PrepareForTest({ LedgerMetaDataCommand.class, MetadataDrivers.class, LedgerMetadata.class,
-    LedgerManagerFactory.class })
+@SuppressWarnings("unchecked")
 public class LedgerMetaDataCommandTest extends BookieCommandTestBase {
 
     private LedgerManager ledgerManager;
     private LedgerManagerFactory factory;
-    private LedgerMetadataSerDe serDe;
-
-    @Mock
     private CompletableFuture<Versioned<LedgerMetadata>> future;
 
-    private TemporaryFolder folder = new TemporaryFolder();
 
     public LedgerMetaDataCommandTest() {
         super(3, 0);
@@ -73,18 +57,9 @@ public class LedgerMetaDataCommandTest extends BookieCommandTestBase {
 
     @Override
     public void setup() throws Exception {
-        folder.create();
-
         factory = mock(LedgerManagerFactory.class);
 
-        PowerMockito.mockStatic(MetadataDrivers.class);
-        PowerMockito.doAnswer(invocationOnMock -> {
-            Function<LedgerManagerFactory, ?> function = invocationOnMock.getArgument(1);
-            function.apply(factory);
-            return true;
-        }).when(MetadataDrivers.class, "runFunctionWithLedgerManagerFactory", any(ServerConfiguration.class),
-                any(Function.class));
-
+        mockMetadataDriversWithLedgerManagerFactory(factory);
         ledgerManager = mock(LedgerManager.class);
         when(factory.newLedgerManager()).thenReturn(ledgerManager);
 
@@ -92,35 +67,39 @@ public class LedgerMetaDataCommandTest extends BookieCommandTestBase {
         when(ledgerMetadata.getMetadataFormatVersion()).thenReturn(1);
         Versioned<LedgerMetadata> versioned = new Versioned<LedgerMetadata>(ledgerMetadata, Version.NEW);
         versioned.setValue(ledgerMetadata);
+        CompletableFuture<Versioned<LedgerMetadata>> future = mock(CompletableFuture.class);
         when(future.join()).thenReturn(versioned);
         when(ledgerManager.readLedgerMetadata(anyLong())).thenReturn(future);
         when(ledgerManager.readLedgerMetadata(anyLong())).thenReturn(future);
         when(future.get()).thenReturn(versioned);
 
-        serDe = mock(LedgerMetadataSerDe.class);
-        whenNew(LedgerMetadataSerDe.class).withNoArguments().thenReturn(serDe);
-        when(serDe.serialize(eq(ledgerMetadata))).thenReturn(new byte[0]);
-        when(serDe.parseConfig(eq(new byte[0]), anyLong(), eq(Optional.empty()))).thenReturn(ledgerMetadata);
+        mockConstruction(LedgerMetadataSerDe.class, (serDe, context) -> {
+            when(serDe.serialize(eq(ledgerMetadata))).thenReturn(new byte[0]);
+            when(serDe.parseConfig(eq(new byte[0]), anyLong(), eq(Optional.empty()))).thenReturn(ledgerMetadata);
+        });
+
         when(ledgerManager.createLedgerMetadata(anyLong(), eq(ledgerMetadata))).thenReturn(future);
     }
 
     @Test
     public void testWithDumpToFile() throws IOException {
-        File file = folder.newFile("testdump");
+        File file = testDir.newFile("testdump");
         LedgerMetaDataCommand cmd = new LedgerMetaDataCommand();
         Assert.assertTrue(cmd.apply(bkFlags, new String[] { "-l", "1", "-d", file.getAbsolutePath() }));
 
         verify(ledgerManager, times(1)).readLedgerMetadata(anyLong());
-        verify(serDe, times(1)).serialize(any(LedgerMetadata.class));
+        verify(getMockedConstruction(LedgerMetadataSerDe.class).constructed().get(0),
+                times(1)).serialize(any(LedgerMetadata.class));
     }
 
     @Test
     public void testWithRestoreFromFile() throws IOException {
-        File file = folder.newFile("testrestore");
+        File file = testDir.newFile("testrestore");
         LedgerMetaDataCommand cmd = new LedgerMetaDataCommand();
         Assert.assertTrue(cmd.apply(bkFlags, new String[] { "-l", "1", "-r", file.getAbsolutePath() }));
 
-        verify(serDe, times(1)).parseConfig(eq(new byte[0]), anyLong(), eq(Optional.empty()));
+        verify(getMockedConstruction(LedgerMetadataSerDe.class).constructed().get(0),
+                times(1)).parseConfig(eq(new byte[0]), anyLong(), eq(Optional.empty()));
         verify(ledgerManager, times(1)).createLedgerMetadata(anyLong(), any(LedgerMetadata.class));
     }
 
